{test program at the end of this unit}
UNIT BINDEX;
INTERFACE
uses opstring;
const
    KEY_LENGTH = 80;            {space for KEY, assume max is 100 char}
    NULL = -1;                   {non-existent page number}
    PAGE_SIZE = 6;               {maximum no. of items in B-tree page}
    HALF_FULL = 3;               {minimum no. of items in B-tree page}

type
    TKEY = string[KEY_LENGTH];
    links = LONGINT;             {B-tree page numbers, or NULL}

    (*
     * "KEYREC" is used to store all the information needed about a KEY
     *  except for the KEY, which is the B-tree search key.
     *)
    TKEYREC = record
	TDATA: TKEY; {RESERVE FOR FURTHER USES}
	lPOS : LONGINT ; {POINTER TO MAIN DATA FILE}
	GAP : STRING[25]; {ALSO RESERVE FOR FURTHER USES}
    end;

    (*
     * "items" are the basic units of the B-tree pages. They have a key
     *  (the KEY), a right child link, and any other information that
     *  is associated with the key (here all the other information
     *  about a KEY).
     *)
    items = record
	KEY: TKEY;
	XDATA: TKEYREC;
	right: links;
    end;

    (*
     * The B-tree consists of a number of pages.
     *  "number" indicates the number of items currently in the page.
     *  Since there needs to be one more child link than there are key
     *   values stored, each page needs one extra link, for the left-most
     *   child.
     *)
    pages = record
	number: 0..PAGE_SIZE;
	left: links;
	item: array[1..PAGE_SIZE] of items;
    end;

    btree_files = file of pages; {the B-tree file structure}

    (*
     * Rather than using a number of global variables, or passing around
     *  a very large number of parameters, all the information associated
     *  with the B-tree file is grouped together into one structure.
     *)
    btrees = record
	master: btree_files;     {the B-tree file itself}
	free_list: links;        {previously-deleted pages list}
	final_record: links;     {highest page number ever used}
	root: links;             {current page number of root page}
    end;

 TBINDEX = OBJECT
   btree: btrees;           {info about the B-tree}
   real_key_len : byte;
   procedure set_key_len(l:byte);
   CONSTRUCTOR INITIALIZE(fname:string);
   DESTRUCTOR TERMINATE;
   procedure add_KEY(Item: items);
   FUNCTION delete_KEY(KEY: TKEY ):BOOLEAN;
   FUNCTION FIND(KEY: TKEY; VAR ITEM : ITEMS):BOOLEAN;
   procedure list_file(branch_no: links);
   procedure summary(branch_no: links; level: integer);
 END;

IMPLEMENTATION

procedure TBINDEX.set_key_len(l:byte);
begin
real_key_len := l;
end;

    (*
     * Name:    random_read
     * Purpose: To input a given page from the B-tree file.
     * Passed:  f:   the file to read
     *          pos: the record number of the page to be read
     * Returns: rec: the page read from the file
     * Notes:   The first record in the file is record 0
     *)
    procedure random_read(var f: btree_files; var rec: pages; pos: links);
    begin
        seek(f, pos);
        read(f, rec);
    end;

    (*
     * Name:    random_write
     * Purpose: To output a given page to the B-tree file.
     * Passed:  f:   the file to write
     *          rec: the page to be written to the file
     *          pos: the record number of the page to be written
     * Notes:   The first record in the file is record 0
     *)
    procedure random_write(var f: btree_files; var rec: pages; pos: links);
    begin
        seek(f, pos);
        write(f, rec);
    end;

    (*
     * Name:    flush
     * Purpose: To output critical information about the current status
     *           of the B-tree to the first record in the file.
     * Passed:  btree: information about the B-tree
     * Notes:   The first record in the B-tree file (record 0) is reserved
     *           for storing the page number of the root node, the page
     *           number of the first page in the free list, and the page
     *           number of the highest page ever used.
     *          This information is stored in the first three links of the
     *           page: left, item[1].right, and item[2].right. This implies
     *           that the page size must be at least 2.
     *)
    procedure flush(var btree: btrees);
    var
        header: pages;
    begin
        header.left := btree.free_list;
	header.item[1].right := btree.final_record;
        header.item[2].right := btree.root;
        random_write(btree.master, header, 0);
    end;

    (*
     * Name:    alloc_page
     * Purpose: To allocate a new page number to become part of the
     *           B-tree.
     * Passed:  btree: information about the B-tree
     * Returns: the number of a suitable page
     * Notes:   This function first tries to find a page that has
     *           previously been deleted. If one exists, it is removed
     *           from the free list, and its number returned.
     *          Failing this, the file must grow by one page.
     *)
    function alloc_page(var btree: btrees): links;
    var
        free_page: pages;
    begin
	if btree.free_list = NULL then begin
            btree.final_record := btree.final_record + 1;
            alloc_page := btree.final_record;
        end else begin
            alloc_page := btree.free_list;
            random_read(btree.master, free_page, btree.free_list);
            btree.free_list := free_page.left;
        end;
    end;

    (*
     * Name:    free_page
     * Purpose: To free a page number that is no longer needed, so that
     *           it can be re-allocated next time a new page is needed.
     * Passed:  btree: information about the B-tree
     *          page:  the number of the page to be added to the free list
     *)
    procedure free_page(var btree: btrees; page: links);
    var
        free_page: pages;
    begin
        (*
         * This page will become the first page on the free list (the
         *  free list is maintained as a stack). The page that used to
         *  be the start of the free list must come next.
         *)
        free_page.left := btree.free_list;
        random_write(btree.master, free_page, page);
        btree.free_list := page;
    end;

    (*
     * Name:    list_one
     * Purpose: To list all the info about one KEY.
     * Passed:  item: information about the KEY
     * Output:  all the info
     *)
    procedure list_one(item: items);
    begin
	writeln('KEY:      ', item.KEY);
	writeln('RET TYPE:    ', item.XDATA.Tdata);
    end;


    (*
     * Name:    initialise
     * Purpose: To open the B-tree file, and read the first record to set
     *           up all the info about free list, root page and final
     *           page in file.
     * Returns: btree: information about the B-tree
     * Notes:   see "flush" for the reciprocal operation at the end of
     *           a session.
     *)
    CONSTRUCTOR TBINDEX.initialiZe;
    var
	header: pages;
    begin
	assign(btree.master, FNAME);
{$I-}
	reset(btree.master);
{$I+}
IF IORESULT <> 0 THEN
 begin
	assign(btree.master, FNAME);
	rewrite(btree.master);
	reset(btree.master);
	btree.free_list := NULL; {no deleted pages yet}
	btree.final_record := 0; {record 0 is used to store status info}
	btree.root := NULL;      {no root yet}
 end
else
begin
	random_read(btree.master, header, 0);
	btree.free_list := header.left;
	btree.final_record := header.item[1].right;
	btree.root := header.item[2].right;
end;
end;

    (*
     * Name:    split
     * Purpose: To split a page that has grown too large.
     * Passed:  btree:     information about the B-tree
     *          item:      the item to be added to the current page
     *          branch_no: the page number of the current page
     *          branch:    the current page
     *          position:  the position within the current page for item
     * Returns: item:      the "middle" item to be promoted to the parent
     *          branch:    current page, now only half full
     * Notes:   To make clear what is happening, I first build an oversize
     *           page with all the items in it.
     *          The bottom half of this big page will remain in the current
     *           page.
     *          The middle item from the big page will be promoted to the
     *           parent.
     *          The top half of the big page will be moved into a new page.
     *)
    procedure split(var btree: btrees; var item: items; branch_no: links;
            var branch: pages; position: integer);
    type
	big_pages = record
            item: array[1..PAGE_SIZE+1] of items;
            number: integer;
        end;
    var
        count: integer;          {for moving items around}
	middle: integer;         {the number of the item to be promoted}
        big_page: big_pages;     {the oversize page}
        new_page: pages;         {the newly created extra page}
        size: integer;           {the number of items in the new page}
    begin
        (*
         * create the page I would like to have...
         *)
        for count := 1 to position-1 do
            big_page.item[count] := branch.item[count];
        big_page.item[position] := item;
        for count := position to branch.number do
            big_page.item[count + 1] := branch.item[count];
        big_page.number := branch.number + 1;

        (*
         * set up the item to be promoted to the parent
         *)
        middle := PAGE_SIZE div 2 + 1;   {item number to promote}
        item := big_page.item[middle];   {item to be promoted}
	item.right := alloc_page(btree); {right child will be new page}

        (*
         * set up the old page
         *)
        for count := 1 to middle - 1 do
            branch.item[count] := big_page.item[count];
        branch.number := middle - 1;

        (*
         * set up the new page...
         *)
        new_page.left := big_page.item[middle].right;
        size := 0;
	for count := middle + 1 to big_page.number do begin
            size := size + 1;
            new_page.item[size] := big_page.item[count];
        end;
        new_page.number := size;

        (*
         * output the modified pages
         *)
        random_write(btree.master, new_page, item.right);
        random_write(btree.master, branch, branch_no);
    end;

    (*
     * Name:    insert
     * Purpose: To insert an item into the current page.
     * Passed:  btree:     information about the B-tree
     *          item:      the item to be added to the current page
     *          branch_no: the page number of the current page
     *          branch:    the current page
     *          index:     the position within the current page for item
     * Returns: item:      the item to be promoted if a split occurred
     *          branch:    current page, as modified
     *          grown:     TRUE if (sub-)tree has grown higher
     *)
    procedure insert(var btree: btrees; var item: items; branch_no: links;
            var branch: pages; index: integer; var grown: boolean);
    var
        count: integer;
    begin
        if branch.number < PAGE_SIZE then begin
            (*
             * room for new item in this page, so no problems
             *)
            grown := FALSE;
            for count := branch.number downto index do
                branch.item[count + 1] := branch.item[count];
            branch.number := branch.number + 1;
            branch.item[index] := item;
            random_write(btree.master, branch, branch_no);
	end else begin
            (*
             * no room here, so split the page...
             *)
            split(btree, item, branch_no, branch, index);
        end;
    end;

    (*
     * Name:    add
     * Purpose: To add an item somewhere in the current "branch" of the
     *           B-tree.
     * Passed:  btree:     information about the B-tree
     *          item:      the item to be added to the current page
     *          branch_no: the page number of the current page
     * Returns: item:      the item to be promoted if a split occurred
     *          branch:    current page, as modified
     *          grown:     TRUE if (sub-)tree has grown higher
     *)
    procedure add(var btree: btrees; var item: items; branch_no: links;
	    var grown: boolean);
    var
        branch: pages;           {the current page}
        index: integer;          {the position for item within branch}
        done: boolean;           {position for item has been found}
        found: boolean;          {exact match has been found}
    begin
        if branch_no = NULL then begin
            (*
             * I have reached beyond the leaf of the tree, so add item
             *  into the previous level.
             *)
            grown := TRUE;
        end else begin
            random_read(btree.master, branch, branch_no);
	    if item.KEY > branch.item[branch.number].KEY then begin
                (*
                 * the new item belongs after the final item in this page
                 *)
                index := branch.number + 1;
		add(btree, item, branch.item[branch.number].right, grown);
                if grown then
                    insert(btree, item, branch_no, branch, index, grown);
            end else begin
                (*
                 * scan the items in this page for the correct position
                 *  for the new item
                 *)
                index := 1;
                found := FALSE;
                done := FALSE;
                while not done and not found do begin
		    if item.KEY < branch.item[index].KEY then
                        done := TRUE
                    else begin
			if item.KEY = branch.item[index].KEY then
                            found := TRUE
                        else
                            index := index + 1;
                    end;
		end;
                if found then begin
		    writeln('ERROR: KEYalready exists');
                    (*
                     * The following statement has been inserted
		     * to make sure a KEYwith duplicate KEY
                     * will not be added when control is returned
		     * back to add_KEYprocedure.
                     *)
		    grown:= FALSE
                end else begin
                    if index = 1 then
                        add(btree, item, branch.left, grown)
                    else
                        add(btree, item, branch.item[index-1].right, grown);
                    if grown then
                        insert(btree, item, branch_no, branch, index, grown);
                end;
            end;
        end;
    end;

    (*
     * Name:    add_KEY
     * Purpose: To add a new KEYinto the library
     *)
    procedure TBINDEX.add_KEY(Item: items);
    var
	{ITEM     = item to be added}
	grown: boolean;          {has the entire B-tree grown higher?}
	page: pages;             {new root page}
	page_no: links;          {page number of new root page}
    begin
	item.right := NULL;
        add(btree, item, btree.root, grown);
        if grown then begin
            (*
             * This growth must be handled as a special case, because the
             *  root is allowed to be less than half full.
             *)
	    page_no := alloc_page(btree);
            page.number := 1;
            page.left := btree.root;
            page.item[1] := item;
            random_write(btree.master, page, page_no);
            btree.root := page_no;
            flush(btree);
        end;
    end;

    (*
     * Name:    balance_from_right
     * Purpose: To "borrow" extra items from a right sibling, to even out
     *           the distribution.
     * Passed:  btree:        information about the B-tree
     *          right_extras: the number of items available to the right
     *          position:     the position in parent separating siblings
     *          child_no:     the page number of the current page
     *          child:        the current page
     *          right_no:     the page number of the right sibling
     *          right:        the right sibling
     *          parent_no:    the page number of the parent page
     *          parent:       the parent page
     * Returns: shrunk:       TRUE if child is undersized
     *          child:        any changes
     *          right:        any changes
     *          parent:       any changes
     *)
    procedure balance_from_right(var btree: btrees; right_extras: integer;
            position: integer; child_no: links; var child: pages;
            right_no: links; var right: pages; parent_no: links;
            var parent: pages; var shrunk: boolean);
    var
        i: integer;
    begin
        (*
         * copy all the items that must be moved from the right sibling
         *  into the current page (child already has the item from the
         *  parent), and adjust the size of the page.
         *)
	for i := 1 to right_extras - 1 do
            child.item[child.number+i] := right.item[i];
        child.number := HALF_FULL + right_extras - 1;
        shrunk := FALSE;

        (*
         * The largest "extra" item from the right sibling becomes
         *  the new parent.
         *)
        parent.item[position] := right.item[right_extras];
        parent.item[position].right := right_no;

        (*
         * fix up right sibling to shift everything left
         *)
        right.left := right.item[right_extras].right;
        right.number := right.number - right_extras;
        for i := 1 to right.number do
            right.item[i] := right.item[right_extras + i];

	(*
         * output the modified pages
         *)
        random_write(btree.master, child, child_no);
        random_write(btree.master, right, right_no);
        random_write(btree.master, parent, parent_no);
    end;

    (*
     * Name:    balance_from_left
     * Purpose: To "borrow" extra items from a left sibling, to even out
     *           the distribution.
     * Passed:  btree:        information about the B-tree
     *          position:     the position in parent separating siblings
     *          child_no:     the page number of the current page
     *          child:        the current page
     *          left_no:      the page number of the left sibling
     *          left:         the left sibling
     *          parent_no:    the page number of the parent page
     *          parent:       the parent page
     *          left_extras:  the number of items available to the left
     * Returns: shrunk:       TRUE if child is undersized
     *          child:        any changes
     *          left:         any changes
     *          parent:       any changes
     *)
    procedure balance_from_left(var btree: btrees; position: integer;
            child_no: links; var child: pages; left_no: links;
            var left: pages; parent_no: links; var parent: pages;
            left_extras: integer; var shrunk: boolean);
    var
        i: integer;
    begin
        (*
         * make room for items to come from left sibling
         *)
        for i := HALF_FULL - 1 downto 1 do
            child.item[i + left_extras] := child.item[i];

        (*
	 * insert the separating item from the parent
         *)
        child.item[left_extras] := parent.item[position];
        child.item[left_extras].right := child.left;

        (*
         * The left sibling has now lost all but one of the "extras".
         *  The final extra will become the new separating item in the
         *  parent.
        left.number := left.number - left_extras + 1;

        (*
         * move the "extras" from the left sibling to the current page
         *)
        for i := left_extras - 1 downto 1 do
            child.item[i] := left.item[left.number + i];
        child.left := left.item[left.number].right;
        child.number := HALF_FULL - 1 + left_extras;
        shrunk := FALSE;

	(*
         * promote the first "extra" to the parent, and remove from left
         *  sibling
         *)
        parent.item[position] := left.item[left.number];
        parent.item[position].right := child_no;
        left.number := left.number - 1;

        (*
         * output the modified pages
         *)
        random_write(btree.master, child, child_no);
        random_write(btree.master, left, left_no);
        random_write(btree.master, parent, parent_no);
    end;

    (*
     * Name:    merge_with_right
     * Purpose: To merge the current page with its right sibling, thus
     *           removing the right sibling from the B-tree.
     * Passed:  btree:        information about the B-tree
     *          position:     the position in parent separating siblings
     *          right_no:     the page number of the right sibling
     *          right:        the right sibling
     *          child_no:     the page number of the current page
     *          child:        the current page
     *          parent_no:    the page number of the parent page
     *          parent:       the parent page
     * Returns: shrunk:       TRUE if child is undersized
     *          child:        any changes
     *          parent:       any changes
     *)
    procedure merge_with_right(var btree: btrees; position: integer;
            right_no: links; right: pages; child_no: links; var child: pages;
            parent_no: links; var parent: pages; var shrunk: boolean);
    var
        i: integer;
    begin
        (*
         * The current page is one less that half full, and the right
	 *  sibling is exactly half full.
         * Add all the items from the right sibling page into the
         *  appropriate places in the current page.
         *)
        for i := 1 to HALF_FULL do
            child.item[HALF_FULL + i] := right.item[i];
        child.number := PAGE_SIZE;

        (*
         * shuffle parent items down to delete the separating item (the
         *  separating item has already been placed in the current page)
         * This may cause the parent page to be undersized...
         *)
        for i := position to parent.number - 1 do
            parent.item[i] := parent.item[i+1];
        parent.number := parent.number - 1;
        shrunk := parent.number < HALF_FULL;

        (*
         * output the modified pages
	 *)
        random_write(btree.master, child, child_no);
        random_write(btree.master, parent, parent_no);

        (*
         * the right sibling can be re-used...
         *)
        free_page(btree, right_no);
    end;

    (*
     * Name:    merge_with_left
     * Purpose: To merge the current page with its left sibling, thus
     *           removing the left sibling from the B-tree.
     * Passed:  btree:        information about the B-tree
     *          position:     the position in parent separating siblings
     *          left_no:      the page number of the left sibling
     *          left:         the left sibling
     *          child_no:     the page number of the current page
     *          child:        the current page
     *          parent_no:    the page number of the parent page
     *          parent:       the parent page
     * Returns: shrunk:       TRUE if child is undersized
     *          left:         any changes
     *          parent:       any changes
     * Notes:   The current page must be the right-most child of the
     *           parent page.
     *)
    procedure merge_with_left(var btree: btrees; position: integer;
            left_no: links; var left: pages; child_no: links; child:
            pages; parent_no: links; var parent: pages; var shrunk: boolean);
    var
        i: integer;
    begin
        (*
         * move the separating item from the parent into the left sibling.
         *)
        left.number := left.number + 1;
        left.item[left.number] := parent.item[position];
        left.item[left.number].right := child.left;

        (*
         * move all the items from the current page into the left sibling
         *)
        for i := 1 to child.number do
            left.item[left.number + i] := child.item[i];
        left.number := PAGE_SIZE;

        (*
         * remove the separating item from the parent. This is very simple,
         *  since merging is always done with the right sibling unless
         *  there is no right sibling, in which case the current page must
         *  be the right-most child of the parent page.
         *)
        parent.number := parent.number - 1;
        shrunk := parent.number < HALF_FULL;

        (*
         * output the modified pages
         *)
	random_write(btree.master, left, left_no);
        random_write(btree.master, parent, parent_no);

        (*
         * the current node can now be re-allocated
         *)
        free_page(btree, child_no);
    end;

    (*
     * Name:    under_flow
     * Purpose: To remedy the situation where the current page has too few
     *           items in it. This may be done either by borrowing items
     *           from a sibling page, or by merging two pages.
     * Passed:  btree:        information about the B-tree
     *          parent_no:    the page number of the parent page
     *          child_no:     the page number of the current page
     *          position:     the position in parent separating siblings
     * Returns: shrunk:       TRUE if child is undersized
     * Notes:   The current page must be the right-most child of the
     *           parent page.
     *)
    procedure under_flow(var btree: btrees; parent_no: links;
            child_no: links; position: integer; var shrunk: boolean);
    var
        child: pages;            {current page}
        parent: pages;           {parent page}
        right_no: links;         {page number of right sibling}
        right: pages;            {right sibling}
        right_extras: integer;   {extra items in right sibling}
        left_no: links;          {page number of left sibling}
        left: pages;             {left sibling}
        left_extras: integer;    {extra items in left sibling}
    begin
        (*
         * load up the relevant pages (maybe already in memory?)
         *)
        random_read(btree.master, child, child_no);
        random_read(btree.master, parent, parent_no);

	if position < parent.number then begin
            (*
             * There is a right sibling.
             *
             * The position in the parent page separating a page from its
             *  right sibling is one greater than the position of the page.
             *)
            position := position + 1;

            (*
             * work out and load the right sibling page
             *)
            right_no := parent.item[position].right;
            random_read(btree.master, right, right_no);

            (*
             * work out how many items should be moved from the right
             *  sibling to the current page to make them even
             *)
            right_extras := (right.number - HALF_FULL + 1) div 2;

            (*
             * regardless of whether items are borrowed from the right
             *  sibling, or the current page is merged with the right
             *  sibling, the current page has the separating item from
             *  the parent added to it.
             *)
            child.number := child.number + 1;
            child.item[child.number] := parent.item[position];
            child.item[child.number].right := right.left;

            if right_extras > 0 then
                (*
                 * borrow some items from the right sibling
                 *)
                balance_from_right(btree, right_extras, position, child_no,
                        child, right_no, right, parent_no, parent, shrunk)
            else begin
                (*
                 * right sibling had nothing to spare
		 *)
                if position <> NULL then begin
                    (*
                     * there is a left sibling, so try borrowing from it
                     *)
                    if position = 1 then
                        left_no := parent.left
                    else
                        left_no := parent.item[position-1].right;
                    random_read(btree.master, left, left_no);
                    left_extras := (left.number - HALF_FULL + 1) div 2;
                    if left_extras > 0 then
                        (*
                         * extras can be borrowed from left sibling
                         *)
                        balance_from_left(btree, position, child_no, child,
                                left_no, left, parent_no, parent,
                                left_extras, shrunk)
                    else
                        (*
			 * left sibling had none to spare either, so give
                         *  up and merge with the right sibling
                         *)
                        merge_with_right(btree, position, right_no, right,
                                child_no, child, parent_no, parent, shrunk);
                end else begin
                    (*
                     * there is no left sibling, so the only option left
                     *  is to merge with the right sibling
                     *)
                    merge_with_right(btree, position, right_no, right,
                            child_no, child, parent_no, parent, shrunk);
                end;
            end;
        end else begin
            (*
             * there is no right sibling, so consider the left
             *)
            if position = 1 then
                left_no := parent.left
	    else
                left_no := parent.item[position-1].right;
            random_read(btree.master, left, left_no);
            left_extras := (left.number - HALF_FULL + 1) div 2;
            if left_extras > 0 then
                (*
                 * left sibling has items to spare
                 *)
                balance_from_left(btree, position, child_no, child, left_no,
                        left, parent_no, parent, left_extras, shrunk)
            else
                (*
                 * left sibling has none to spare, so the only option is
                 *  to merge with the left sibling.
                 *)
                merge_with_left(btree, position, left_no, left, child_no,
                        child, parent_no, parent, shrunk);
        end;
    end;

    (*
     * Name:    find_predecessor
     * Purpose: To find the leaf item which is the predecessor of the item
     *           to be deleted (branch), swap the item to be deleted
     *           with its predecessor, and then delete the item.
     * Passed:  btree:        information about the B-tree
     *          index:        the position in branch of item
     *          descend_no:   the page number of the descendant page
     *          branch_no:    the page number of the page containing item
     *          branch:       the page containing the item to be deleted
     * Returns: shrunk:       TRUE if descend is undersized
     *          branch:       any changes
     *)
    procedure find_predecessor(var btree: btrees; index: integer;
            descend_no: links; branch_no: links; var branch: pages;
            var shrunk: boolean);
    var
        child: links;            {right-most child of descend page}
        descend: pages;          {current descendant page}
    begin
	(*
	 * load the current page
         *)
        random_read(btree.master, descend, descend_no);

        (*
         * see if there is a right-most child
         *)
        child := descend.item[descend.number].right;
        if child = NULL then begin
            (*
             * this is the end of the line. "descend" is the right-most
             *  descendant to the left of the item being deleted. Since
             *  this is its predecessor, the two items can be swapped
             *  without affecting the rest of the tree. Thus the
             *  descendant is copied into the ancestor, and the extra copy
             *  left of the descendant is then deleted.
             *)
            descend.item[descend.number].right := branch.item[index].right;
            branch.item[index] := descend.item[descend.number];
	    descend.number := descend.number - 1;
            shrunk := descend.number < HALF_FULL;

            (*
             * update the modified pages
             *)
            random_write(btree.master, branch, branch_no);
            random_write(btree.master, descend, descend_no);
        end else begin
            (*
             * more descendants, so keep following them down
             *)
	    find_predecessor(btree, index, child, branch_no, branch, shrunk);

            (*
             * on the way back up, keep on topping up undersized pages as
             *  necessary.
             *)
            if shrunk then
                under_flow(btree, descend_no, child, descend.number, shrunk);
	end;
    end;

    (*
     * Name:    delete
     * Purpose: To delete an item with a given key from the current
     *           sub-tree.
     * Passed:  btree:        information about the B-tree
     *          KEY:         the key to be deleted
     *          branch_no:    the page number of root of the current sub-tree
     * Returns: shrunk:       TRUE if branch is undersized
     *)
    procedure delete(var btree: btrees; KEY: TKEY; branch_no: links;
            var shrunk: boolean);
    var
        branch: pages;           {the root of the current sub-tree}
	index: integer;          {the position in branch of KEY}
        i: integer;              {used in small loops}
	child: pages;            {child of branch that might contain KEY}
        child_no: links;         {page number of child}
	found: boolean;          {has KEY been located in B-tree?}
        done: boolean;           {has correct position been determined?}
    begin
	if branch_no = NULL then begin
	    (*
	     * I have reached the bottom of the tree without finding the
	     *  key I wanted to delete
	     *)
	    shrunk := FALSE;
	end else begin
            (*
             * load in the current page
             *)
            random_read(btree.master, branch, branch_no);

	    if KEY > branch.item[branch.number].KEY then begin
                (*
		 * KEY is to the right of the right-most item, so follow
		 *  the right-most line of descent
                 *)
                index := branch.number;
                child_no := branch.item[branch.number].right;
		delete(btree, KEY, child_no, shrunk);
                if shrunk then
                    under_flow(btree, branch_no, child_no, index, shrunk);
            end else begin
                (*
                 * locate the correct path to descend to the item to be
                 *  deleted
                 *)
                index := 1;
                found := FALSE;
                done := FALSE;
                while not done and not found do begin
		    if KEY < branch.item[index].KEY then
                        done := TRUE
                    else
			if KEY = branch.item[index].KEY then
			    found := TRUE
                        else
                            index := index + 1;
                end;
                if found then begin
                    (*
		     * KEY has been located! Now see if it was on a leaf
		     *  page or not. *)
			if index = 1 then
                            child_no := branch.left
                        else
                            child_no := branch.item[index-1].right;
                        if child_no = NULL then begin
			    (*
                             * on a leaf page already. Delete the item, and
                             *  set the "shrunk" flag if necessary so that
                             *  the next level up can fix the undersize
                             *  problem
                             *)
                            branch.number := branch.number - 1;
                            shrunk := branch.number < HALF_FULL;
                            for i := index to branch.number do
                                branch.item[i] := branch.item[i+1];
                            random_write(btree.master, branch, branch_no)
                        end else begin
                            (*
                             * not on a leaf node. Deletions can only be
			     *  done directly on leaf nodes, so I must
			     *  swap the current item with its predecessor
			     *)
			    find_predecessor(btree, index, child_no,
				    branch_no, branch, shrunk);
			    if shrunk then
				under_flow(btree, branch_no, child_no,
					index-1, shrunk);
			end;
		end else begin
                    (*
                     * not on this page, so keep on descending and
                     *  searching for it.
                     *)
                    if index = 1 then
                        child_no := branch.left
                    else
                        child_no := branch.item[index-1].right;
		    delete(btree, KEY, child_no, shrunk);
                    if shrunk then
                        under_flow(btree, branch_no, child_no, index-1,
                                shrunk);
                end;
            end;
        end;
    end;

    (*
     * Name:    delete_KEY
     * Purpose: To delete a KEYfrom the library database.
     * Passed:  btree:        information about the B-tree
     *)
    FUNCTION TBINDEX.delete_KEY(KEY: TKEY ):BOOLEAN;
    var
	    {key to be deleted}
	root: pages;             {the root page of the B-tree}
	shrunk: boolean;         {is the root undersized?}
	old_root: links;         {page no. of root with no items}
    begin
	(*
	 * find out which KEYis to be deleted
	 *)
	delete(btree, KEY, btree.root, shrunk);
	DELETE_KEY := shrunk;
	if shrunk then begin
	    random_read(btree.master, root, btree.root);
            if root.number = 0 then begin
                (*
                 * The old root has shrunk to the point where it has no
                 *  items and only one left-most child. This left-most
                 *  child must become the new root.
                 * The old root can then be freed.
                 *)
                old_root := btree.root;
                btree.root := root.left;
                flush(btree);
                free_page(btree, old_root);
            end;
        end;
    end;

    (*
     * Name:    locate
     * Purpose: To locate an item with a given key in the current sub-tree.
     * Passed:  btree:        information about the B-tree
     *          KEY:         the key to be deleted
     *          branch_no:    the page number of root of the current sub-tree
     * Returns: page_no:      the number of the page containing KEY
     *          page:         the page containing KEY
     *          index:        the position within page of KEY
     *)
    PROCEDURE locate(var btree: btrees; KEY: TKEY; branch_no: links;
	    var page_no: links; var page: pages; var index: integer);
    var
	branch: pages;           {the root of the current sub-tree}
	i: integer;              {index while searching}
	done: boolean;           {KEY position determined?}
	found: boolean;          {KEY located?}
    begin
	if branch_no = NULL then begin
	    (*
	     * I have reached the bottom of the tree without finding the
	     *  KEY I was looking for.
	     *)
	    page_no := NULL;
	end else begin
	    (*
	     * search the current page...
	     *)
	    random_read(btree.master, branch, branch_no);
            key := opstring.StUpcase(trim(KEY));
            branch.item[branch.number].KEY :=opstring.StUpcase(trim(branch.item[branch.number].KEY));
	    if key > branch.item[branch.number].KEY then begin
		(*
		 * search to right of right-most item
		 *)
		locate(btree, KEY, branch.item[branch.number].right,
			page_no, page, index);
	    end else begin
		(*
		 * find correct sub-tree to search
		 *)
		i := 1;
		found := FALSE;
		done := FALSE;
		while not done and not found do begin
		    if KEY < branch.item[i].KEY then
			done := TRUE
		    else begin
			if KEY = branch.item[i].KEY then
			    found := TRUE
			else
                            i := i + 1;
		    end;
		end;
		if found then begin
		    (*
		     * got it! Now just set up the return parameters
		     *)
		    page := branch;
		    page_no := branch_no;
		    index := i;
		end else begin
		    (*
		     * follow appropriate sub-tree
		     *)
		    if i = 1 then
			locate(btree, KEY, branch.left, page_no, page,
				index)
		    else
			locate(btree, KEY, branch.item[i-1].right,
				page_no, page, index);
		end;
	    end;
	end;
    end;

    (*
     * Name:    locate_KEY
     * Purpose: To display info about a KEY.
     * Passed:  btree: information about the B-tree
     *)
    FUNCTION TBINDEX.FIND;
    var
	     {KEY of borrowed KEY}
	page_no: links;          {page number KEY occurred on, if any}
	page: pages;             {page KEY occurred on}
	index: integer;          {position if KEY within page}
    begin
       FIND := FALSE;
	locate(btree, KEY, btree.root, page_no, page, index);
	if page_no <> NULL then begin
	   FIND := TRUE;
	   ITEM :=page.item[index];
	end;
    end;

    (*
     * Name:    terminate
     * Purpose: To save the current status of the B-tree, and close the
     *           file.
     *)
    DESTRUCTOR TBINDEX.terminate;
    begin
	flush(btree);
        close(btree.master);
    end;

    (*
     * Name:    summary
     * Purpose: To display the keys of all the items in the B-tree, so that
     *          the B-tree algorithms can be checked.
     * Passed:  btree:     info about the B-tree
     *          branch_no: the root of the current sub-tree
     *          level:     the level of the current root in the tree
     *)
    procedure TBINDEX.summary(branch_no: links; level: integer);
    var
        i: integer;              {counter}
        branch: pages;           {root of current sub-tree}
    begin
        if branch_no <> NULL then begin
            (*
             * load the current page
             *)
	    random_read(btree.master, branch, branch_no);

            (*
             * indent by the required amount of space (in proportion to
             *  the level in the tree)
             *)
            for i := 1 to level do
                write('':10);

            (*
	     * display the items in the current page
             *)
            for i := 1 to branch.number do
		write(branch.item[i].KEY:10);
            writeln;

            (*
             * recurse to display all the children of this page
             *)
	    summary(branch.left, level+1);
	    for i := 1 to branch.number do
		summary(branch.item[i].right, level+1);
        end;
    end;


(*function get_most_LEFT(var btree: btrees) : LONGINT;
VAR AREC, arec2 : PAGES;
    I: BYTE;
BEGIN
open_stack;
SEEK(BTREE.MASTER,BTREE.ROOT);
READ(btree.MASTER, AREC);
  FOR I := AREC.NUMBER  downto 1 DO
    add_stack(arec.item[i].right);
WHILE AREC.LEFT > 0 DO
BEGIN
  SEEK(BTREE.MASTER,AREC.LEFT);
  READ(btree.MASTER, AREC);
  FOR I := 1 TO AREC.NUMBER DO
    WRITE(AREC.ITEM[I].KEY, '  ');
END;
  FOR I := 1 TO AREC.NUMBER DO
  if arec.item[i].right > 0 then
   begin

   end;

WRITELN;
END;
*)

    (*
     * Name:    list_file
     * Purpose: To traverse the B-tree and display of all the items
     *          in the B-tree, in ascending order of KEY.
     * Passed:  btree:     info about the B-tree
     *          branch_no: the root of the current sub-tree
     *)

procedure TBINDEX.list_file(branch_no: links);
    var
        i: integer;              {counter}
        branch: pages;           {details of current page}
    begin
        if branch_no <> NULL then begin
	    (*
             * load the current page
             *)
	    random_read(btree.master, branch, branch_no);

            (*
             * recurse to get to the leftmost page
	     *)
	    list_file(branch.left);
            (*
             * process each item on the current page
             *)
            for i := 1 to branch.number do begin
                (*
                 * display each item details
                 *)
                list_one(branch.item[i]);
                writeln;
                (*
                 * recurse to process right child if any
		 *)
		list_file(branch.item[i].right);
            end
	end;
    end;
end.

(***** test program ***)
{copy this into other file to run and test btree unit}
PROGRAM TEST_BINDEX;
USES BINDEX;

 FUNCTION GET_KEY:STRING;
    VAR KEY : STRING;
    BEGIN
    GET_KEY := '';
    write('Enter KEY to be located: ');
    readln(KEY);
    GET_KEY := KEY;
    END;
    (*
     * Name:    input_KEY
     * Purpose: To input all the info about one KEY.
     * Returns: item: information about the KEY
     * Input:   all the info
     *)
  procedure input_KEY(var item: items);
    begin
	write('Enter KEY:      ');
	readln(item.KEY);

	write('Enter RET TYPE:    ');
	readln(item.XDATA.Tdata);
    end;

    (*
     * Name:    list_one
     * Purpose: To list all the info about one KEY.
     * Passed:  item: information about the KEY
     * Output:  all the info
     *)
    procedure list_one(item: items);
    begin
	writeln('KEY:      ', item.KEY);
	writeln('RET TYPE:    ', item.XDATA.Tdata);
    end;

    (*
     * Name:    input_KEY
     * Purpose: To input all the info about one KEY.
     * Returns: item: information about the KEY
     * Input:   all the info
     *)


    (*
     * Name:    menu
     * Purpose: To display a menu, and get the user to input a valid
     *           selection
     * Returns: a valid selection from the menu
     *)
    function menu: char;
    var
        selection: char;         {user's selection}
    begin
        (*
         * keep on displaying the menu and asking for a selection until
         *  the user enters a valid selection.
         *)
        repeat
            writeln;
	    writeln(' Database Menu');
	    writeln;
	    writeln('    A) Add new KEY');
	    writeln('    D) Delete outdated or lost KEY');
	    writeln('    L) Locate KEY');
	    writeln('    P) Print KEYs list in KEY order');
	    writeln;
	    writeln('    Q) Quit this program');
	    writeln;
	    write('Selection: ');
	    readln(selection);
	    selection := UPCASE(selection);
	until selection in ['A', 'D', 'L', 'Q', 'P', '*'];
        menu := selection;
        writeln;
    end;


    (*
     * Name:    main
     * Purpose: To display a menu, get a selection, and call the
     *           appropriate function.
     * Notes:   This is a separate module, rather than the body of the
     *           main program, to avoid the presence of global variables.
     *)
    procedure main;
    var
	finished: boolean;       {user wishes to quit this program?}
	Item: items;
	KEY: TKEY;
	MY : TBINDEX;
	ANSWER : CHAR;
    begin
	(*
	 * restore all the status info about the location of the root,
	 *  the start of the free list, and the number of records in
	 *  the file
	 *)
	MY.initialiZe('client.ndx');

	(*
	 * let the user do as much as desired...
	 *)
	finished := FALSE;
	repeat
	WITH MY DO
	    case menu of
	    'A':BEGIN
		Input_KEY(item);
		add_KEY(ITEM);
		END;
	    'D':
		BEGIN
		KEY := GET_KEY;
		IF FIND(KEY,ITEM) THEN
		BEGIN
		 LIST_ONE(ITEM);
	    write('Delete this KEY? (Y/N): ');
	    readln(answer);
	    if (answer in ['y', 'Y']) then
		 delete_KEY(KEY);
		END;
		END;
	    'L':
		BEGIN
		KEY := GET_KEY;
		IF FIND(KEY,ITEM) THEN
		 LIST_ONE(ITEM);
		END;
	    'P': {get_most_LEFT(BTREE);}
		list_file(MY.BTREE.ROOT);
	    '*':
		summary(btree.root, 0);
	    'Q':
		finished := TRUE;
	    end;
	until finished;

	(*
	 * save the status info for next time
	 *)
	MY.terminate;
    end;
BEGIN
writeln(sizeof(pages));
MAIN;
END.